/*
 * defRead.c --      
 *
 * This module incorporates the LEF/DEF format for standard-cell place and
 * route.
 *
 * Version 0.1 (September 26, 2003):  DEF input of designs.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /ufs/repository/magic/lef/defRead.c,v 1.0 2003/05/01 10:37:00 tim Exp $";            
#endif  /* not lint */

#include <stdio.h>
#include <stdlib.h>
#ifdef SYSV
#include <string.h>
#endif
#include <errno.h>
#include <stdarg.h>

#include "magic/tclmagic.h"
#include "misc/magic.h"
#include "utils/geometry.h"
#include "tiles/tile.h"
#include "utils/hash.h"
#include "database/database.h"
#include "windows/windows.h"
#include "dbwind/dbwind.h"
#include "utils/malloc.h"
#include "graphics/graphics.h"
#include "main/main.h"
#include "cif/cif.h"
#include "lef/lefInt.h"

/*
 *------------------------------------------------------------
 *
 * DefAddRoutes --
 *
 *	Parse a network route statement from the DEF file,
 *	and add it to the linked list representing the route.
 *
 *------------------------------------------------------------
 */

void
DefAddRoutes(lineptr, rlistptr, curLayer, oscale)
    char *lineptr;
    linkedRoute **rlistptr;
    TileType curLayer;
    float oscale;
{
    char *tptr = lineptr;
    char token[256];
    linkedRoute *newRoute, *routeList = *rlistptr;
    TileType newLayer = curLayer;
    float x, y;
    Point p;
   
    /* Go to end of list */
    if (routeList != NULL)
	while (routeList->route_next != NULL)
	    routeList = routeList->route_next;

    while (*tptr != '\0')
    {
	/* Just in case we're not at the beginning of the next token */
        while(isspace(*tptr)) tptr++;

	/* Get next point, or via name */
	if (*tptr != '(')	/* via name */
	{
	    sscanf(tptr, "%255s", token);
	    tptr = LefSkipToken(tptr);
	    if (routeList == NULL)
	    {
		LefError("Route has via name \"%s\" but no points!\n", token);
		continue;
	    }
	    newLayer = DBTechNameType(LefLower(token));
	    if (newLayer >= 0)
	    {
		newRoute = (linkedRoute *)mallocMagic(sizeof(linkedRoute));
		newRoute->type = newLayer;
		newRoute->p = routeList->p;
		newRoute->route_next = NULL;
	    }
	    else
	    {
		LefError("Via name \"%s\" unknown in route.\n", token);
		return;
	    }
	}
	else
	{
	    /* Read an (X Y) point */
	    tptr++;
	    while(isspace(*tptr)) tptr++;
	    if (*tptr == '*')
	    {
		if (routeList == NULL)
		{
		    LefError("No reference point for \"*\" wildcard\n"); 
		    goto endCoord;
		}
		p.p_x = routeList->p.p_x;
		tptr++;
	    }
	    else if (sscanf(tptr, "%f", &x) == 1)
	    {
		p.p_x = (int)(x / oscale);
	        while(!isspace(*tptr)) tptr++;
	    }
	    else
	    {
		LefError("Cannot parse X coordinate.\n"); 
		goto endCoord;
	    }
	    while(isspace(*tptr)) tptr++;
	    if (*tptr == '*')
	    {
		if (routeList == NULL)
		{
		    LefError("No reference point for \"*\" wildcard\n"); 
		    freeMagic(newRoute);
		    newRoute = NULL;
		    goto endCoord;
		}
		p.p_y = routeList->p.p_y;
		tptr++;
	    }
	    else if (sscanf(tptr, "%f", &y) == 1)
	    {
		p.p_y = (int)(y / oscale);
	        while(!isspace(*tptr) && (*tptr != ')')) tptr++;
	    }
	    else
	    {
		LefError("Cannot parse Y coordinate.\n"); 
		goto endCoord;
	    }

	    while(isspace(*tptr)) tptr++;
	    if (*tptr != ')')
	    {
		/* non-default route extension */
		/* (to be done) */
	    }

	    newRoute = (linkedRoute *)mallocMagic(sizeof(linkedRoute));
	    newRoute->type = newLayer;
	    newRoute->route_next = NULL;
	    newRoute->p = p;

endCoord:
	    /* Find the closing parenthesis for the coordinate pair */
	    while(*tptr != ')')
	    {
		if (*tptr == '\0')
		{
		    LefError("Line ended without end of coordinate.\n");
		    return;
		}
		tptr++;
	    }
	    tptr = LefSkipToken(tptr);
	}

	/* Link in the new route segment */
	if (newRoute)
	{
	    if (routeList)
		routeList->route_next = newRoute;
	    else
		*rlistptr = newRoute;
	    routeList = newRoute;
	    newRoute = NULL;
	}
    }
}

/*
 *------------------------------------------------------------
 *
 * DefPaint --
 *
 *	Paint a network route into the layout
 *
 *------------------------------------------------------------
 */

void
DefPaint(routeList, rootDef)
    linkedRoute *routeList;
    CellDef *rootDef;
{
    Point *p1, *p2;
    Rect area, crect;
    TileType rType;
    linkedRoute *rp = routeList;

    if (rp == NULL) return;

    /* Pick first point, then move to the next record and remove. */
    /* This code will handle the case of a single point record	  */
    /* (e.g., "NEW VIA12 ( 0 0 )"				  */

    p1 = &rp->p;
    rType = rp->type;

    while (rp != NULL)
    {
        if (rp->route_next != NULL)	/* pick up 2nd point of segment */
	    p2 = &rp->route_next->p;
	else if (rp->type != rType)	/* via at end of route */
	{
	    p2 = p1;
	    rType = rp->type;
	}
	else				/* no via; nothing more to route */
	    goto freepoint;

	/* compute area to paint */
	area.r_xbot = p1->p_x;
	area.r_ybot = p1->p_y;
	area.r_xtop = p2->p_x;
	area.r_ytop = p2->p_y;
	GeoCanonicalRect(&area, &crect);

	/* This is a placeholder!  Must get routing info from the LEF file! */
	crect.r_xbot -= 1;
	crect.r_xtop += 2;
	crect.r_ybot -= 1;
	crect.r_ytop += 2;

	/* paint */
	DBPaint(rootDef, &crect, rType);

	/* advance to next point and free record (1-delayed) */
	p1 = p2;

freepoint:

	rType = rp->type;
	freeMagic((char *)rp);
	rp = rp->route_next;
    }
}

/*
 *------------------------------------------------------------
 *
 * DefReadNets --
 *
 *	Read a NETS section from a DEF file.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	Many.  Networks are created, and geometry may be
 *	painted into the database top-level cell.
 *
 *------------------------------------------------------------
 */

enum def_net_keys {DEF_NET_START = 0, DEF_NET_PROP, DEF_NET_MORE, 
	DEF_NET_NEW, DEF_NET_END};
enum def_netprop_keys {DEF_NETPROP_USE, DEF_NETPROP_ROUTED, DEF_NETPROP_FIXED,
	DEF_NETPROP_COVER};
enum def_net_expect {DEF_EXPECT_NONE = 0, DEF_EXPECT_NET, DEF_EXPECT_ROUTE};

void
DefReadNets(f, rootDef, sname, oscale)
    FILE *f;
    CellDef *rootDef;
    char *sname;
    float oscale;
{
    char token[256], *tptr;
    char *lineptr;
    int keyword, subkey, total_nets, expect;
    TileType curLayer;
    linkedRoute *routeList;

    static char *net_keys[] = {
	"-",
	"+",
	"(",
	"NEW",
	"END",
	NULL
    };

    static char *net_property_keys[] = {
	"USE",
	"ROUTED",
	"FIXED",
	"COVER",
	NULL
    };

    routeList = NULL;
    expect = DEF_EXPECT_NONE;

    while ((lineptr = LefNextLine(f)) != NULL)
    {
	if (sscanf(lineptr, "%s", token) == 0)
	    continue;   /* Blank line or null statement */

	keyword = Lookup(token, net_keys);
	if (keyword < 0)
	{
	    /* Check for "(" as a continuation line.  It is not clear	*/
	    /* from the spec whether "(" must be followed by whitespace */
	    /* but we assume legal syntax either way.			*/

	    if (token[0] == '(')
		keyword = DEF_NET_MORE;
	    else
	    {
		LefError("Unknown keyword \"%s\" in NET "
			"definition; ignoring.\n", token);
		continue;
	    }
	}
	tptr = lineptr;

	if (keyword != DEF_NET_MORE)
	{
	    /* Process any pending information */
	    DefPaint(routeList, rootDef);
	    routeList = NULL;
	    curLayer = -1;
	}

	switch (keyword)
	{
	    case DEF_NET_START:
		expect = DEF_EXPECT_NET;
		break;
	    case DEF_NET_PROP:
		tptr = LefSkipToken(lineptr);
		if (sscanf(tptr, "%s", token) != 1)
		    LefError("Bad network property\n");
		else
		{
		    subkey = Lookup(token, net_property_keys);
		    if (subkey < 0)
		    {
			LefError("Unknown network property \"%s\" in "
				"NET definition; ignoring.\n", token);
			continue;
		    }
		    switch (subkey)
		    {
			case DEF_NETPROP_USE:
			    expect = DEF_EXPECT_NONE;
			    break;
			case DEF_NETPROP_ROUTED:
			    expect = DEF_EXPECT_ROUTE;
			    break;
			case DEF_NETPROP_FIXED:
			    expect = DEF_EXPECT_ROUTE;
			    break;
			case DEF_NETPROP_COVER:
			    expect = DEF_EXPECT_ROUTE;
			    break;
			default:
			    expect = DEF_EXPECT_NONE;
			    break;
		    }
		}
		if (expect != DEF_EXPECT_ROUTE)
		    break;

		/* drop through to below to get layer type */

	    case DEF_NET_NEW:
		if (expect != DEF_EXPECT_ROUTE)
		{
		    LefError("Statement NEW does not continue a route\n");
		    break;
		}
		tptr = LefSkipToken(tptr);
		sscanf(tptr, "%s", token);
		tptr = LefSkipToken(tptr);

		/* Note that, like lefRead.c, this measure is ad-hoc */

		curLayer = DBTechNameType(LefLower(token));
		if (curLayer < 0)
		{
		    LefError("Unknown layer type \"%s\" for route\n", token); 
		    break;
		}

		/* drop through to below to get route points */

	    case DEF_NET_MORE:
		if (expect == DEF_EXPECT_ROUTE)
		    DefAddRoutes(tptr, &routeList, curLayer, oscale);
		break;
	    case DEF_NET_END:
		if (!LefParseEndStatement(lineptr, sname))
		{
		    LefError("Net END statement missing.\n");
		    keyword = -1;
		}
		break;
	}
	if (keyword == DEF_NET_END) break;
    }
		tptr = LefSkipToken(tptr);

    /* Process any pending information before returning */
    DefPaint(routeList, rootDef);
}

/*
 *------------------------------------------------------------
 *
 * DefReadUseLocation --
 *
 *	Read location and orientation of a cell use
 *	Syntax: ( X Y ) O
 *
 *------------------------------------------------------------
 */
enum def_orient {DEF_NORTH, DEF_SOUTH, DEF_EAST, DEF_WEST,
	DEF_FLIPPED_NORTH, DEF_FLIPPED_SOUTH, DEF_FLIPPED_EAST,
	DEF_FLIPPED_WEST};

void
DefReadLocation(use, lineptr, oscale)
    CellUse *use;
    char *lineptr;
    float oscale;
{
    Rect *r;
    int keyword;
    char token[5];
    float x, y;
    Transform t;

    static char *orientations[] = {
	"N", "S", "E", "W", "FN", "FS", "FE", "FW"
    };

    if (sscanf(lineptr, "( %f %f ) %4s", &x, &y, token) != 3)
    {
	LefError("Badly formed macro orientation and location.\n");
	return;
    }

    keyword = Lookup(token, orientations);
    if (keyword < 0)
    {
	LefError("Unknown macro orientation \"%s\".\n", token);
	return;
    }

    /* The standard transformations are all defined to rotate	*/
    /* or flip about the origin.  However, DEF defines them	*/
    /* around the center such that the lower left corner is	*/
    /* unchanged after the transformation.  Case conditions	*/
    /* restore the lower-left corner position.			*/

    r = &use->cu_def->cd_bbox;
    switch (keyword)
    {
	case DEF_NORTH:
	    t = GeoIdentityTransform;
	    t.t_c = -(r->r_xbot);
	    t.t_f = -(r->r_ybot);
	    break;
	case DEF_SOUTH:
	    t = Geo180Transform;
	    t.t_c = r->r_xtop;
	    t.t_f = r->r_ytop;
	    break;
	case DEF_EAST:
	    t = Geo90Transform;
	    t.t_c = -(r->r_ybot);
	    t.t_f = r->r_xtop;
	    break;
	case DEF_WEST:
	    t = Geo270Transform;
	    t.t_c = r->r_ytop;
	    t.t_f = -(r->r_xbot);
	    break;
	case DEF_FLIPPED_NORTH:
	    t = GeoSidewaysTransform;
	    t.t_c = r->r_xtop;
	    t.t_f = -(r->r_ybot);
	    break;
	case DEF_FLIPPED_SOUTH:
	    t = GeoUpsideDownTransform;
	    t.t_c = -(r->r_xbot);
	    t.t_f = r->r_ytop;
	    break;
	case DEF_FLIPPED_EAST:
	    t = GeoRef135Transform;
	    t.t_c = r->r_ytop;
	    t.t_f = r->r_xtop;
	    break;
	case DEF_FLIPPED_WEST:
	    t = GeoRef45Transform;
	    t.t_c = -(r->r_ybot);
	    t.t_f = -(r->r_xbot);
	    break;
    }

    /* Translate --- see GeoTranslateTrans() in utils/geometry.c */
    /* t_c -> x translation, t_f -> y translation */

    t.t_c += (int)(x / oscale);
    t.t_f += (int)(y / oscale);

    DBSetTrans(use, &t);
}

/*
 *------------------------------------------------------------
 *
 * DefReadComponents --
 *
 *	Read a COMPONENTS section from a DEF file.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	Many.  Cell definitions and uses are created and added to
 *	the database.
 *
 *------------------------------------------------------------
 */

enum def_comp_keys {DEF_COMP_START = 0, DEF_COMP_PROP, DEF_COMP_END};
enum def_prop_keys {DEF_PROP_FIXED = 0, DEF_PROP_COVER, DEF_PROP_PLACED,
	DEF_PROP_UNPLACED};

void
DefReadComponents(f, rootDef, sname, oscale)
    FILE *f;
    CellDef *rootDef;
    char *sname;
    float oscale;
{
    CellDef *defMacro;
    CellUse *defUse;
    char token[256], *tptr;
    char *lineptr;
    char defname[256];
    int keyword, subkey, total_uses = 0;

    static char *component_keys[] = {
	"-",
	"+",
	"END",
	NULL
    };

    static char *property_keys[] = {
	"FIXED",
	"COVER",
	"PLACED",
	"UNPLACED",
	NULL
    };

    defUse = NULL;
    while ((lineptr = LefNextLine(f)) != NULL)
    {
	if (sscanf(lineptr, "%s", token) == 0)
	    continue;   /* Blank line or null statement */
	keyword = Lookup(token, component_keys);

	if (keyword < 0)
	{
	    LefError("Unknown keyword \"%s\" in COMPONENT "
			"definition; ignoring.\n", token);
	    continue;
	}
	switch (keyword)
	{
	    case DEF_COMP_START:
		if (defUse != NULL)	/* Finish previous call, if any */
		    DBPlaceCell(defUse, rootDef);

		defMacro = NULL;	/* Forget the last def */
		defUse = NULL;		/* Forget the last use */
		tptr = LefSkipToken(lineptr);
		if (sscanf(tptr, "%s %s", token, defname) != 2)
		    LefError("Bad component statement:  Need use and macro names\n");
		else
		{
		    /* Find the corresponding macro definition */
		    defMacro = DBCellLookDef(defname);
		    if (defMacro == (CellDef *)NULL)
		    {
			LefError("Cell %s is not defined.  Maybe you have not "
				"read the corresponding LEF file?\n",
				defname);
			break;
		    }

		    /* Create a use for this celldef in the edit cell */
		    defUse = DBCellNewUse(defMacro, token);
		    total_uses++;
		    if (total_uses % 1000 == 0)
			TxPrintf("  Processed %d subcell instances.\n", total_uses);
		}
		break;
	    case DEF_COMP_PROP:
		
		/* Don't process properties for cells we could not find */
		if (defUse == NULL) break;

		tptr = LefSkipToken(lineptr);
		if (sscanf(tptr, "%s", token) != 1)
		    LefError("Bad component property\n");
		else
		{
		    subkey = Lookup(token, property_keys);
		    if (subkey < 0)
		    {
			LefError("Unknown component property \"%s\" in "
				"COMPONENT definition; ignoring.\n", token);
			continue;
		    }
		    switch (subkey)
		    {
			case DEF_PROP_PLACED:
			case DEF_PROP_UNPLACED:
			case DEF_PROP_FIXED:
			case DEF_PROP_COVER:
			    tptr = LefSkipToken(tptr);
			    DefReadLocation(defUse, tptr, oscale);
			    break;
		    }
		}
		break;
	    case DEF_COMP_END:
		if (!LefParseEndStatement(lineptr, sname))
		{
		    LefError("Component END statement missing.\n");
		    keyword = -1;
		}

		/* Finish final call by placing the cell use */
		if (defUse != NULL)
		    DBPlaceCell(defUse, rootDef);

		break;
	}
	if (keyword == DEF_COMP_END) break;
    }
    TxPrintf("  Processed %d subcell instances total.\n", total_uses);
}

/*
 *------------------------------------------------------------
 *
 * DefRead --
 *
 *	Read a .def file into a magic layout.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	Many.  Cell definitions and uses are created and added to
 *	the database.
 *
 *------------------------------------------------------------
 */

enum def_sections {DEF_VERSION = 0, DEF_NAMESCASESENSITIVE,
	DEF_UNITS, DEF_DESIGN, DEF_ROW, DEF_TRACKS,
	DEF_GCELLGRID, DEF_DIVIDERCHAR, DEF_BUSBITCHARS,
	DEF_PROPERTYDEFINITIONS, DEF_DEFAULTCAP,
	DEF_HISTORY, DEF_DIEAREA, DEF_COMPONENTS, DEF_VIAS,
	DEF_PINS, DEF_PINPROPERTIES, DEF_SPECIALNETS,
	DEF_NETS, DEF_IOTIMINGS, DEF_SCANCHAINS,
	DEF_CONSTRAINTS, DEF_GROUPS, DEF_EXTENSION,
	DEF_END};

void
DefRead(inName)
    char *inName;
{
    CellDef *rootDef;
    FILE *f;
    char *filename;
    char token[256];
    char *lineptr;
    int keyword, dscale;
    float oscale;

    static char *sections[] = {
	"VERSION",
	"NAMESCASESENSITIVE",
	"UNITS",
	"DESIGN",
	"ROW",
	"TRACKS",
	"GCELLGRID",
	"DIVIDERCHAR",
	"BUSBITCHARS",
	"PROPERTYDEFINITIONS",
	"DEFAULTCAP",
	"HISTORY",
	"DIEAREA",
	"COMPONENTS",
	"VIAS",
	"PINS",
	"PINPROPERTIES",
	"SPECIALNETS",
	"NETS",
	"IOTIMINGS",
	"SCANCHAINS",
	"CONSTRAINTS",
	"GROUPS",
	"BEGINEXT",
	"END",
	NULL
    };

    f = lefFileOpen(NULL, inName, ".def", "r", &filename);

    if (f == NULL)
    {
#ifdef MAGIC_WRAPPER
	TxError("Cannot open input file %s (%s).\n", filename,
		strerror(errno));
#else
	TxError("Cannot open input file: ");
	perror(filename);
#endif
	return;
    }

    /* Initialize */

    TxPrintf("Reading DEF data from file %s.\n", filename);
    TxPrintf("This action is undoable.\n");
    UndoDisable();

    /* This works for CIF reads;  maybe should only do this if the top	*/
    /* cell is (UNNAMED)?						*/

    rootDef = EditCellUse->cu_def;
    DBCellRenameDef(rootDef, inName);
    oscale = CIFGetOutputScale(1000);
    lefCurrentLine = 0;

    /* Read file contents */

    while ((lineptr = LefNextLine(f)) != NULL)
    {
	if (sscanf(lineptr, "%s", token) == 0)
	    continue;	/* Blank line or null statement */

	keyword = Lookup(token, sections);
	if (keyword < 0)
	{
	    LefError("Unknown keyword \"%s\" in DEF file; ignoring.\n", token);
	    continue;
	}
	switch (keyword)
	{
	    case DEF_VERSION:
		break;
	    case DEF_NAMESCASESENSITIVE:
		break;
	    case DEF_DESIGN:
		break;
	    case DEF_UNITS:
		if (sscanf(lineptr, "%*s %*s %*s %d", &dscale) != 1)
		{
		    LefError("Invalid syntax for UNITS statement.\n");
		    LefError("Assuming default value of 100\n");
		    dscale = 100;
		}
		/* We don't care if the scale is 100, 200, 1000, or 2000. */
		/* Do we need to deal with numeric roundoff issues?	  */
		oscale *= (float)dscale;
		break;
	    case DEF_ROW:
		break;
	    case DEF_TRACKS:
		break;
	    case DEF_GCELLGRID:
		break;
	    case DEF_DIVIDERCHAR:
		break;
	    case DEF_BUSBITCHARS:
		break;
	    case DEF_HISTORY:
		break;
	    case DEF_DIEAREA:
		break;
	    case DEF_PROPERTYDEFINITIONS:
		LefSkipSection(f, sections[DEF_PROPERTYDEFINITIONS]);
		break;
	    case DEF_DEFAULTCAP:
		LefSkipSection(f, sections[DEF_DEFAULTCAP]);
		break;
	    case DEF_COMPONENTS:
		DefReadComponents(f, rootDef, sections[DEF_COMPONENTS], oscale);
		break;
	    case DEF_VIAS:
		LefSkipSection(f, sections[DEF_VIAS]);
		break;
	    case DEF_PINS:
		LefSkipSection(f, sections[DEF_PINS]);
		break;
	    case DEF_PINPROPERTIES:
		LefSkipSection(f, sections[DEF_PINPROPERTIES]);
		break;
	    case DEF_SPECIALNETS:
		LefSkipSection(f, sections[DEF_SPECIALNETS]);
		break;
	    case DEF_NETS:
		DefReadNets(f, rootDef, sections[DEF_NETS], oscale);
		break;
	    case DEF_IOTIMINGS:
		LefSkipSection(f, sections[DEF_IOTIMINGS]);
		break;
	    case DEF_SCANCHAINS:
		LefSkipSection(f, sections[DEF_SCANCHAINS]);
		break;
	    case DEF_CONSTRAINTS:
		LefSkipSection(f, sections[DEF_CONSTRAINTS]);
		break;
	    case DEF_GROUPS:
		LefSkipSection(f, sections[DEF_GROUPS]);
		break;
	    case DEF_EXTENSION:
		LefSkipSection(f, sections[DEF_EXTENSION]);
		break;
	    case DEF_END:
		if (!LefParseEndStatement(lineptr, "DESIGN"))
		{
		    LefError("END statement out of context.\n");
		    keyword = -1;
		}
		break;
	}
	if (keyword == DEF_END) break;
    }
    TxPrintf("DEF read:  Done reading file\n");
    LefError(NULL);	/* print statement of errors, if any, and reset */

    /* Cleanup */

    DBAdjustLabels(rootDef, &TiPlaneRect);
    DBReComputeBbox(rootDef);
    DBWAreaChanged(rootDef, &rootDef->cd_bbox, DBW_ALLWINDOWS,
		&DBAllButSpaceBits);
    DBCellSetModified(rootDef, TRUE);

    UndoDisable();
    if (f != NULL) fclose(f);
}
